/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

/* check.c
 * 
 *  check for open socks server
 */

#include "config.h"

#include <strings.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/time.h>
#include "extern.h"
#include "configuration.h"
#include "defines.h"

int debug;
int mode=1;
int verbose=0;
int dbon=0;
int port=0;
int allow_kill=1;

struct passwd *userinfo;

	/*
	 * Not run as CGI so Parse command line params:
	 * -h <hostname>
	 * -v verbose mode (verbose about connections and results - no debug info)  [optional]
	 * -p <port> [optional]
	 * -a paranoid mode (reject if socks found) [optional]
	 * -r relaxed mode (reject on open socks only) [optional]
	 * -n normal mode (default, reject on authenticated or open proxy) [optional]
	 * -w write to db mode (write ip address to db based upon detection level) [optional]
	 * -d <debug level> [optional]
	 */

main(int argc, char **argv) {

	int i;
	int r;
	int t;
	int sock=0;

	char *hostname;

#ifdef DEBUG
	 debug=5;
#else
	 debug=0;
#endif
	if (!parse_config(argc, argv)) usage(argv[0]);
	
	IFDEBUG( { \
		printf("Debug Level=%i\n", debug); \
		print_debug(2, "Hostname: %s\n", config->hostname?config->hostname:"not specified"); \
		print_debug(2, "Ports: 1080, 3128, 808%s%i\n", port?"0, ":"", port); \
		print_debug(2, "Mode: %s\n", mode?mode==2?"Relaxed":"Normal":"Paranoid"); \
		print_debug(2, "DB creation: %s\n", dbon?"on":"off"); \
		print_config(); \
	} );

	if (!config->hostname) usage(argv[0]);

	/* Get uid of this process, and set userinfo pw entry */
	get_userinfo();

	if (config->mode==DAEMON) {
		start_daemon_mode();
	}

	if (config->mode==STANDALONE) {
		for(i=0;i<config->ports;i++) {
			switch (config->port[i]->type) {
				case HTTP:
					t=0;
					while (!r && config->testhost[t]) {
						r=test_proxy(config->hostname, config->port[i]->port, config->testhost[t]->hostname, config->testhost[t]->port);
						if (r) {
							reject(config->exitval->http);
							break;
						}
						t++;
					}
					break;
				case SOCKS:
					r=test_sock(config->hostname, config->port[i]->port);
					if (r) reject(config->exitval->socks);
					break;
				case SMTP:
					r=test_relay(config->hostname);
					if (r) reject(config->exitval->relay);
					break;
				default:
					break;
			}
		}
	}

	cleanup();
	
}

void get_userinfo(void) {

	userinfo=getpwuid(geteuid());
	if (!userinfo) {
		perror("Unable to get userinfo");
		exit (1);
	}
}

int usage(char *progname) {
	fprintf(stderr, "Usage: %s [-v] [-r] [-n] [-a] [-w] -h <hostname> [-p <port>]\r\n", progname);
	exit(1);
}

void reject(int exit_val) {

	int t;
	unsigned char *i;

	i=getip(config->hostname);
	for(t=0; t<4; t++, i++) {
		printf("%hhi%s", *i, t<3?".":"");
	}
	printf(": rejected after tests\n");
	cleanup();
	exit(exit_val);

}

char *unescape_string(char *s) {

	char *p, *pt, *t, buf[3];
	int l, ret;

	p=s;
	l=strlen(s);
	l++;
	pt=t=calloc(sizeof(char), l);

	while (*s) {
		if (*s == '%') {
			s++; // found % we want the next 2 chars
			buf[0]=*s++;
			buf[1]=*s++;
			buf[2]='\0';
			ret=(int) strtol(buf, NULL, 16);
			*t++=ret;
		} else {
			*t++=*s++;
		}
	}
	t=pt;
	s=p;
	while (*t) *s++=*t++;
	*s='\0';
	free(pt);
}
