/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif


#include <stdio.h>
#include <errno.h>
#include <signal.h>

#include "defines.h"
#include "extern.h"

/* signal.c - a collection of functions to setup and handle signals.... */

int setup_sighandler(void) {

	struct sigaction sigact;

	sigact.sa_handler = SIG_IGN;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags = 0;
	if (sigaction(SIGTERM, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGTERM failed.\n");
	if (sigaction(SIGHUP, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGHUP failed.\n");
	if (sigaction(SIGINT, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGINT failed.\n");

	sigact.sa_handler = do_signal;
	sigemptyset(&sigact.sa_mask);
	sigact.sa_flags |= SA_RESTART;

	if (sigaction(SIGTERM, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGTERM failed.\n");
	if (sigaction(SIGHUP, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGHUP failed.\n");
	if (sigaction(SIGINT, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGINT failed.\n");
	if (sigaction(SIGCHLD, &sigact, (void *) NULL)) fprintf(stderr, "sigaction(): Setting SIGCHLD failed.\n");

}

void do_signal(int sig) {
	
	switch (sig) {
		case SIGSEGV:
		case SIGTERM:
		case SIGINT:
			printf("Caught %s.  Exiting....\n", sig==SIGSEGV?"SIGSEGV":sig==SIGTERM?"SIGTERM":"SIGINT");
			cleanup();
			exit(0);
			break; // we'll never hit this!
		case SIGCHLD:
			catch_child(sig);
			break;
		default:
			printf("Ignoring Signal [%d]\n", sig);
			break;
	}

}

void catch_child(int sig) {

	int cstatus;

	wait(&cstatus);
	IFDEBUG(print_debug(3, "Child process exited...\n"));
}

