/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif


#include <pthread.h>

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>

#include "defines.h"
#include "dbconfig.h"
#include "extern.h"

/* 
 * declared in db.c
 *
 * struct conn_struct connection[MAX_REQUESTS];
 * struct dbreq_struct req[MAX_REQUESTS];
 * int request[MAX_REQUESTS];
 * int thread_started;
 *
 * declared in dbconfig.h
 * struct t_args {
 * int id;
 * int sd;
 * char *ipaddr;
 * char *hostname;
 * };
 */

void *new_dbthread(void *);
void *new_client_thread(void *);
void do_request(int);
void reset_req(int, int);

/* start_client_thread():
 *
 * needs: open and accepted socket
 *        incoming ipaddress in as a string
 *        incoming hostname as a string.
 *
 * returns: 
 *
 *
 *
 */
int start_client_thread(int sd, char *ipaddr, char *hostname) {

	int i=0;
	int error=0;
	struct t_args *tap;


	pthread_attr_t detached_attr;

	char *t;
	t=malloc(32768);
	memset(t, 0, 32768);
	free(t);
	 
	pthread_attr_init(&detached_attr);
	pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);

	i=find_free_conn();
	if (i<0) {
		net_send(sd, "Too many connections, try later\n");
		fprintf(stderr, "Rejecting connection from %s, there are already %i connections.\n", hostname, MAX_REQUESTS);
		close(sd);
		return -1;
	}

	tap = (struct t_args *)calloc(sizeof(struct t_args), 1);
	tap->id=i;
	tap->sd=sd;
	tap->ipaddr=(char *)strdup(ipaddr);
	free(ipaddr);
	tap->hostname=hostname;

	thread_started=0;

	if (pthread_create(&connection[i].tid, &detached_attr, new_client_thread, (void *) tap)) {
		fprintf(stderr, "Error creating ");
		perror("connection thread");
		return -1;
	}

	/* don't need else - if there was an error we have already returned to the parent */
	while(!thread_started) usleep(10);

}

int find_free_conn() {
	
	int i;

	for (i=0; i<MAX_REQUESTS; i++) {
		if (connection[i].status==FREE) return i;
	}
	return -1;
}

void start_threads(void) {

	int i;
	int error=0;
	struct t_args *tap;

	pthread_attr_t detached_attr;
	
	pthread_attr_init(&detached_attr);
	pthread_attr_setdetachstate(&detached_attr, PTHREAD_CREATE_DETACHED);

	for(i=0; i<MAX_REQUESTS; i++) {

		tap = (struct t_args *)calloc(sizeof(struct t_args), 1);
		
		connection[i].status=FREE;
		request[i]=0;
		tap->id=i;
		if (pthread_create(&connection[i].tid, &detached_attr, new_dbthread, (void *) tap)) {
			fprintf(stderr, "Error creating ");
			perror("thread");
			error=1;
		}

		if (error) {
			fprintf(stderr, "Error creating threads on startup - Quitting....\n");
			cleanup();
			exit(1);
		}
	}
}

void *new_dbthread(void *tap) {

	/* ok thread started, now it has to wait for an incoming request
	 * these will be communicated via a global structure.
	 *
	 * id is my thread number for request tracking.
	 *
	 */
	int id;
	struct t_args *arg = (void *)tap;

	id=arg->id;
	free(arg);

	init_db_threads(); /* this is to initialise thread safe calls fdr the DB clients
											* for example in mysqldb.c it called mysql_thread_init();
											*
											* This function is found in db.c
											*
											*/

	req[id].status=IDLE;
	req[id].query=NONE;
	req[id].result=NONE;
	req[id].client=NULL;

	while (1) {

		while(req[id].status==IDLE || req[id].status==DONE) sleep(1);
		switch(req[id].client->type) {
			case CHECK:
				req[id].status=BUSY;
				req[id].result=do_dbcheck(id);
				req[id].status=DONE;
				break;
			case ADD:
				req[id].status=BUSY;
				req[id].result=do_dbadd(id);
				req[id].status=IDLE;
				break;
			case DEL:
				req[id].status=BUSY;
				req[id].result=do_dbdel(id);
				req[id].status=IDLE;
				break;
			case EXIT:
				req[id].status=BUSY;
#ifdef HAVE_MYSQL
				disconnect_db_thread();
#endif
				req[id].status=GONE;
				pthread_exit(0);
				break; /* never gonna get here ... */
			default:
				break;
		}
	}
}

void *new_client_thread(void *tap) {

	/* ok this is a new client thread started and we have the socket it's listening to... */

	int id, len;
	char *hostname;
	char *ipaddr = NULL;
	char *buf = NULL;
	char c, t, *addrp;
	char msg[256];
	int sd, ret;
	int port=0;
	struct t_args *arg = (void *)tap;
	char type[] = "EHRSTWOLD";

	id=arg->id;
	connection[id].status=BUSY; /* we don't want anything else starting as thread[id] */
	connection[id].result=NONE;
	connection[id].query=NONE;

	sd=arg->sd;
	ipaddr=arg->ipaddr;
	hostname=arg->hostname;
	free(arg);
	thread_started++; /* this will release the parent */

	// kill the thread
	IFDEBUG(print_debug(3, "Thread[%i]: Accepted Connection...\n", id));
	if (net_send(sd, "200 Database Ready\n")<0) goto exit;
	while(1) {
		buf=net_read(sd);
		if (!buf || *buf == 'Q') {
			if (!buf) {
				net_send(sd, "400 Timeout, closing connection...\n");
				fprintf(stderr, "Thread[%i]: Read timeout from host %s [%s] quitting...\n", id, hostname, ipaddr);
			} else {
				net_send(sd, "220 bye\n");
				IFDEBUG(fprintf(stderr, "Thread[%i]: %s [%s] quit.\n", id, hostname, ipaddr));
			}
			goto exit;
		}
		
		len=strlen(buf);
		len++;
		addrp=(char *)malloc(len);
		memset(addrp, '\0', len);

		if (buf) {
			switch(*buf) {
				case 'C':
					sscanf(buf, "%c %s", &c, addrp);
					break;
				case 'D':
					sscanf(buf, "%c %c %s", &c, &t, addrp);
					break;
				case 'A':
					sscanf(buf, "%c %c %s %i", &c, &t, addrp, &port);
					connection[id].port=port;
					break;
				default:
					fprintf(stderr, "Thread[%i]: ERROR! Command=%c\n", id, *buf);
					break;
			}
		}

		len=strlen(addrp);
		connection[id].type=CHECK;
		memset(connection[id].ipaddr, '\0', 17);
		strncpy(connection[id].ipaddr, addrp, len<16?len:16);
		connection[id].result=NONE;

		ret=0;
	
		if (c=='C') goto skip;

		switch (t) {
			case 'H':
				connection[id].query=HTTP;
				break;
			case 'S':
				connection[id].query=SOCKS;
				break;
			case 'R':
				connection[id].query=SMTP;
				break;
			case 'T':
				connection[id].query=TRAP;
				break;
			case 'W':
				connection[id].query=WHITE;
				break;
			case 'O':
				connection[id].query=OKLIST;
				break;
			case 'D':
				connection[id].query=RDNS;
				break;
			default:
				ret=-1;
				fprintf(stderr, "Query %c [%hhi] not recognised from client %s\n", t, t, ipaddr);
				break;
		}
skip:
		if (!ret) {
			switch(c) {
				case 'C':
					start_checking(id);
					ret=get_results(id);
					break;
				case 'A':
					ret=do_dbadd(setup_req(id));
					// reset_req(id,0);
					break;
				case 'D':
					ret=do_dbdel(setup_req(id));
					// reset_req(id,0);
					break;
				default:
					ret=-1;
					fprintf(stderr, "Command %c [%hhi] not recognised from client %s\n", c, c, ipaddr);
					break;
			}
		}
		if (!ret) {
			memset(msg, '\0', 256);
			sprintf(msg, "250 %s Not found in DB\n", connection[id].ipaddr);
			if (net_send(sd, msg)<0) goto exit;
		} else {
			if (ret<0) {
				if (net_send(sd, "500 Command not recognised\n")<0) goto exit;
			} else {
				IFDEBUG(print_debug(5, "Connection-Thread[%i]: Result=%i\n", id, ret));
				memset(msg, '\0', 256);
				switch (ret) {
					case DELPROB:
						sprintf(msg, "45%i %s Host was NOT deleted\n", connection[id].query, connection[id].ipaddr);
						break;
					case ADDPROB:
						sprintf(msg, "45%i %s Host was NOT added\n", connection[id].query, connection[id].ipaddr);
						break;
					case DELETED:
						sprintf(msg, "25%i %s Host deleted\n", connection[id].query, connection[id].ipaddr);
						break;
					case ADDED:
						sprintf(msg, "25%i %s Host added\n", connection[id].query, connection[id].ipaddr);
						break;
					default:
						switch (connection[id].query) {
							case HTTP:
							case SMTP:
							case SOCKS:
							case TRAP:
								sprintf(msg, "45%i %s This is an open server\n", connection[id].query, connection[id].ipaddr);
								break;
							case WHITE:
							case OKLIST:
							case LOCALLIST:
								sprintf(msg, "25%i %s This is a known good host\n", connection[id].query, connection[id].ipaddr);
								break;
							default:
								sprintf(msg, "55%i %s An error Occured\n", connection[id].query, connection[id].ipaddr);
								break;
						}
						break;
				}
				if (net_send(sd, msg)<0) goto exit;
				reset_req(id, 0);
			}
		}
		free(buf);
		free(addrp);
		buf=NULL; /* safety */
		addrp=NULL; /* safety */
	}
exit:
	if (buf) free(buf);
	if (addrp) free(addrp);
	close(sd);
	connection[id].status=FREE;
	pthread_exit(0);
}

void reset_req(int id, int n) {

	int i;
	for(i=0; i<n; i++) {
		req[connection[id].dbthread[i]].result=NONE;
		memset(req[connection[id].dbthread[i]].ipaddr, '\0', 16);
		req[connection[id].dbthread[i]].status=FREE;
		req[connection[id].dbthread[i]].type=NONE;
		req[connection[id].dbthread[i]].query=NONE;
		if (req[connection[id].dbthread[i]].hostname) {
			free(req[connection[id].dbthread[i]].hostname);
			req[connection[id].dbthread[i]].hostname=NULL;
		}
		req[connection[id].dbthread[i]].client=NULL;
		connection[id].dbthread[i]=-1;
	req[connection[id].dbthread[i]].type=NONE;
	}
}

int setup_req(id) {

	int dbtid;

	dbtid=-1;
	while(dbtid<0) {
		dbtid=find_free_dbthread();
		usleep(10);
	}
	req[dbtid].client=&connection[id];
	connection[id].dbthread[0]=dbtid; /* tracking of which threads we used */
	return dbtid;
}

void cleanup_threads() {

	int i;

	for (i=0; i<MAX_REQUESTS; i++) {
		while (req[i].status != FREE) {
			if (req[i].status == GONE) i++; /* catch seemingly impossible */
			sleep(1);
		}
		req[i].query=EXIT;
	}
}

int find_free_dbthread(void) {

	int i;

	for (i=0; i<MAX_REQUESTS; i++) {
		if (req[i].status==FREE) {
			IFDEBUG(print_debug(5, "Selecting Thread: %i for list processing\n", i));
			return i;
		}
	}
	return -1;
}

int start_checking(int id) {

	int t;
	int dbtid;

	for (t=0; t<7; t++) {
		
		dbtid=-1;
		while(dbtid<0) {
			dbtid=find_free_dbthread();
			usleep(10);
		}
		req[dbtid].client=&connection[id];
		connection[id].dbthread[t]=dbtid; /* tracking of which threads we used */
		switch(t) {
			case 0:
				req[dbtid].query=OKLIST;
				break;
			case 1:
				req[dbtid].query=WHITE;
				break;
			case 2:
				req[dbtid].query=SMTP;
				break;
			case 3:
				req[dbtid].query=HTTP;
				break;
			case 4:
				req[dbtid].query=SOCKS;
				break;
			case 5:
				req[dbtid].query=TRAP;
				break;
			case 6:
				req[dbtid].query=RDNS;
				break;
		}
		req[dbtid].status=GO;
	}
}

int get_results(int id) {

	int t=0;
	int ret=0;
	int res=0;
	
	ret=connection[id].result=NOT_MATCHED;

	for (t=0; t<7; t++) {
		while(req[connection[id].dbthread[t]].status!=DONE) usleep(100);
		IFDEBUG(print_debug(5, "Connection-Thread[%i]: Checking Result for DB-Thread[%i]\n", id, connection[id].dbthread[t]));
		if (req[connection[id].dbthread[t]].result==MATCHED) {
			IFDEBUG(print_debug(5, "Connection-Thread[%i]: Got Matched Result for DB-Thread[%i]\n", id, connection[id].dbthread[t]));
			connection[id].query=req[connection[id].dbthread[t]].query;
			connection[id].result=MATCHED;

			/* We want the first result - however we should wait for all the threads to return */
			if (!res) {
				ret=connection[id].result;
				res++;
			}
		}
	}

	/* reset all threads */
	for (t=0; t<7; t++) {
		req[connection[id].dbthread[t]].result=NONE;
		req[connection[id].dbthread[t]].status=FREE;
		req[connection[id].dbthread[t]].client=NULL;
		req[connection[id].dbthread[t]].query=NONE;
		connection[id].dbthread[t]=-1;
	}
	return ret;
}

