/* Debug Header */
#if DEBUG
# define IFDEBUG(x) do { if (debug) x; }while(0)
# define IFDEBUG5(x) do { if (debug) x; }while(0)
#else
# define IFDEBUG(x)
# define IFDEBUG5(x)
#endif
/* End Debug Header */


#if HAVE_CONFIG_H
# include <config.h>
#endif

#include "config.h"

#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include "defines.h"
#include "configuration.h"
#include "extern.h"

void wait_for_quit(char *, int);

void dealloc_results(void) {

	int i;

	for(i=0; i<config->ports; i++) {
		free(config->results[i]);
	}
	free(config->results);

}

void alloc_results(void) {

	int i;

	config->results=(struct test_struct **)calloc(sizeof(struct test_struct *), config->ports);
	for (i=0; i<config->ports; i++) {
		config->results[i]=(struct test_struct *)calloc(sizeof(struct test_struct), 1);
	}

}

void do_connection(char *ipaddr) {

	int i;
	char *buf;
	int res=0;
	int match=0;
	int fd;
	int serverd=0;

	/* ok we have just fored off after acceptng a connection on a socket */

	allow_kill=0;

	match=check_db_match(ipaddr);
	IFDEBUG(print_debug(3, "Client[]: DB returned %i for %s\n", match, ipaddr));
	if (match==250 || match<0) {
		IFDEBUG(print_debug(3, "Client[]: Starting test sequence\n"));
		/* a code of 250 indicates no match - we do checking
		 * a code of -1 indicates a dtabase error - we do checking
		 */
		alloc_results();
	
		/* spin_lock will do as it says.... spin until it gets the lock
	 	* 
	 	* ... the lock is a common lock specified in the config or a
	 	* #define'd default.
	 	*
	 	* This is need to ensure that the other processes are just waiting
	 	* on results.  This cures both a race condition and a deadlock.
	 	*/
		fd=spin_lock(config->lockfile?config->lockfile:SEM_LOCK);
		/* Now we run the tests */
		start_tests(config->hostname);
		/* all tests are running (results unknown as yet, but we can safely release the lock */
		release_lock(fd);
	
		res=check_results(config->hostname, ipaddr);
		if (res && config->mode) {
			match=res+450;
		} else {
			if(config->listgood) add_addr(OKLIST, ipaddr, 0); /* if list good is not specified will retest each time */
		}

		dealloc_results();
	}
	buf=(char *)calloc(sizeof(char), NET_BUFFER+1);
	if (match>400 || (res && config->thread->type)) {
		IFDEBUG(print_debug(3, "Client[]: Bad Host - closing connection\n"));
		/* matched bad host */
		if (config->strictsmtp) {
			sprintf(buf, "554 %s\n", config->smtpmsg?config->smtpmsg:"You are in my banned hosts list");
			write(config->thread->sockfd, buf, strlen(buf)+1);
			wait_for_quit(buf, config->thread->sockfd);
		} else {
			sprintf(buf, "421 %s\n", config->smtpmsg?config->smtpmsg:"You are in my banned hosts list");
			write(config->thread->sockfd, buf, strlen(buf)+1);
		}
		
	} else {
		/* matched good/whitelisted host */
		IFDEBUG(print_debug(3, "Client[]: Good Host - connecting to backend...\n"));
		serverd=net_connect(config->targethost?config->targethost:"localhost", config->targetport?config->targetport:25);
		do_thru_connect(serverd, config->thread->sockfd);
		close(serverd);

	}

	close(config->thread->sockfd);
	free(ipaddr);
	free(buf);
	cleanup();

}

void wait_for_quit(char *buf, int sfd) {
	
	char *str=NULL;

	while(1) {
		str=net_read(sfd);
		memset(buf, '\0', NET_BUFFER+1);
		if (str) {
			if ((* str   =='Q' || * str   =='q') &&
			    (*(str+1)=='U' || *(str+1)=='u') &&
			    (*(str+2)=='I' || *(str+2)=='i') &&
			    (*(str+3)=='T' || *(str+3)=='t')) {
				sprintf(buf, "250 Bye\n");
				net_send(sfd, buf);
				break;
			} else {
				sprintf(buf, "503 Command out of sequence\n");
				net_send(sfd, buf);
				free(str);
				continue;
			}
		}
		sprintf(buf, "421 I've waited long enough, goodbye\n");
		net_send(sfd, buf);
		break;
	}
	free(str);
}

			
int check_results(const char *hostname, const char *ipaddr) {

	char buf[256];
	int p=0;
	int *t;
	int i;
	int ret=0;

	IFDEBUG(print_debug(5, "[Parent] In check_results()...\n"));
	for(i=0; i<config->ports; i++) {
		while(check_free(config->results[i]->id) != OK) sleep(5);
		t=&config->results[i]->type;
		/* anything but O is bad */
		memset(buf, '\0', 256);
		sprintf(buf, "R");
		write(config->workers[config->results[i]->id]->childfd, buf, strlen(buf)+1);
		IFDEBUG(print_debug(5, "[Parent] Requesting results from worker [%i]\n", i));
		config->results[i]->result=read_child_pipe(config->workers[config->results[i]->id]->parentfd, config->workers[config->results[i]->id]->id, 'R', t, IDLE);
		if (reset_worker(config->results[i]->id)) {
			IFDEBUG(print_debug(5, "[Parent] Reset Worker[%i] to IDLE\n", i));
		} else {
			IFDEBUG(print_debug(5, "[Parent] Error Reseting Worker[%i] to IDLE\n", i));
		}
		IFDEBUG(print_debug(5, "[Parent] Result for %s test is %s for host: %s on port: %i\n", \
				config->results[i]->type?config->results[i]->type==SMTP?"SMTP": \
				config->results[i]->type==HTTP?"HTTP":"SOCKS":"IDLE",  \
				config->results[i]->result?"bad":"good", hostname, config->results[i]->port));
			
		if (config->results[i]->result) {
			ret=config->results[i]->type;
			add_addr(ret, ipaddr, config->results[i]->port);
		}
	}
	return ret;
}

void start_tests(const char *hostname) {

	int i, t, s, c;

	c=config->children;
	i=0;
	t=0;
	s=0;

	while(c--) {

		while (check_free(i)) {
			i++;
			if (i==config->children) {
				i=0;
				sleep(5);
			}
		}
		send_command(hostname, c, i);
		config->results[c]->id=i; /* set the thread id for the result */
		config->results[c]->port=config->port[c]->port;
	}

}

int reset_worker(int i) {

	char buf[256];
	int s;
	int cid=0;
	int *pcid;

	pcid=&cid;

	memset(buf, '\0', 256);
	sprintf(buf, "T %i %s 0", IDLE, "localhost", 0);
	write(config->workers[i]->childfd, buf, strlen(buf)+1);
	return(read_child_pipe(config->workers[i]->parentfd, config->workers[i]->id, 'T', pcid, BUSY));
	
}

int check_free(int i) {

	char buf[256];
	int s;
	int cid=0;
	int *pcid;

	pcid=&cid;

	memset(buf, '\0', 256);
	sprintf(buf, "S %i", config->workers[i]->id);
	write(config->workers[i]->childfd, buf, strlen(buf)+1);
	return(read_child_pipe(config->workers[i]->parentfd, config->workers[i]->id, 'S', pcid, BUSY));
	
}

int read_child_pipe(int sd, int id, int c, int *cid, int ret) {

	int r=0;
	fd_set rfds;
	struct timeval tv;
	char buf[256];
	char command[256];

	FD_ZERO(&rfds);
	FD_SET(sd, &rfds);
	tv.tv_sec = PIPE_TIMEOUT;
	tv.tv_usec = 0;

	r=select(sd+1, &rfds, NULL, NULL, &tv);
	IFDEBUG({ \
		if (r)  \
			print_debug(3, "[Parent] Parent<-Thread[%i]: Pipe has data.\n", id); \
		else { \
			print_debug(3, "[Parent] Parent<-Thread[%i]: Pipe has no data before timeout.\n", id); \
		}
	});
	if (!r)	goto out;

	memset(buf, '\0', 256);
	memset(command, '\0', 256);
	r=read(sd, buf, 255);
	IFDEBUG(print_debug(4, "[Parent] Read: %s from Thread[%i]\n", buf, id));
	sscanf(buf, "%c %i %i", command, cid, &ret);
	if (*command != 'R' && *cid != id) {
		fprintf(stderr, "[Parent] Error: Pipe for Thread[%i] returned [%c] ID=%i\n", id, *command, *cid);
		ret=BUSY;
		goto out;
	}
	IFDEBUG(print_debug(3, "[Parent] Thread[%i]: %c=%i\n", id, *command, ret));

out:
	return ret;

}

void send_command(const char *hostname, int c, int id) {

	int i=0;
	int ns=0;
	char buf[256];
	int *pcid;
	int cid=0;

	pcid=&cid;

	memset(buf, '\0', 256);
	sprintf(buf, "T %i %s %i", config->port[c]->type, hostname, config->port[c]->port);
	write(config->workers[id]->childfd, buf, strlen(buf)+1);
	if (read_child_pipe(config->workers[id]->parentfd, config->workers[id]->id, 'T', pcid, BUSY))
		IFDEBUG(print_debug(3, "[Parent] Thread[%i]: Checking '%s:%i'\n", config->workers[id]->id, hostname, config->port[c]->port));
		else fprintf(stderr, "[Parent] Error: Thread[%i] Did not accept the target comand\n", config->workers[id]->id);

}
